#! /usr/bin/env python3 -u
# coding=utf-8

# -----------------------------------------------------------------------------
# Copyright 2021 Martin Jan Köhler
# See http://creativecommons.org/licenses/MIT/ for more information.
# -----------------------------------------------------------------------------

import argparse
from ctypes import *
import csv
import datetime
import difflib
from enum import *
import inspect
import json
import os
import pathlib
import platform
import re
import serial
import statistics
import struct
import sys
import time

from scipy import stats

# -----------------------------------------------------------------------------
# NOTE: python<3.5 does not yet have bytes.hex()
pyver = list(map(lambda n: int(n), platform.python_version_tuple()))
if pyver[0] < 3 or pyver[1] < 5:
    raise Exception("Requires Python version >= 3.5")


# -----------------------------------------------------------------------------

def prompt(msg):
    return "%s%s%s> " % (term_color_green, msg, term_color_reset)


def print_error(msg):
    print("%sERROR:%s %s\n" % (term_color_red, term_color_reset, msg))


def print_attention(msg):
    print("%sATTENTION: %s %s" % (term_color_yellow, msg, term_color_reset))


# -----------------------------------------------------------------------------

isSaveMode = False
is_open_mode = False


def complete(text, state):
    global isSaveMode
    global is_open_mode
    if isSaveMode or is_open_mode:
        if text == "" or text is None:
            text = "."
        p = pathlib.Path(text)

        options = []
        for subpath in p.parent.glob("%s*" % p.name):
            options.append(str(subpath))

        return options[state]
    else:
        return None


complete('', 0)

# -----------------------------------------------------------------------------

try:
    import readline
    import rlcompleter

    if platform.system() == 'Darwin':
        readline.parse_and_bind("bind ^I rl_complete")
    else:
        readline.parse_and_bind('tab: complete')
    readline.parse_and_bind('tab: complete')
    readline.set_completer(complete)
    readline.set_completer_delims("\t")

    term_color_green = ''
    term_color_red = ''
    term_color_yellow = ''
    term_color_reset = ''

    term_with_colors = ['xterm', 'xterm-color', 'xterm-256color', 'linux', 'screen', 'screen-256color', 'screen-bce']
    if os.environ.get('TERM') in term_with_colors:
        term_color_green = '\033[32m'
        term_color_red = '\033[31m'
        term_color_yellow = '\033[33m'
        term_color_reset = '\033[0m'
except ImportError:
    pass

# -----------------------------------------------------------------------------

isVoltmeterSupported = True
try:
    import voltmeter
except ImportError:
    isVoltmeterSupported = False


# -----------------------------------------------------------------------------

def convert(d, transformers):
    for k, v in d.items():
        if k in transformers.keys():
            d[k] = (transformers[k])(v)
        else:
            d[k] = convert(d.get(k, {}), transformers)
    return d


def bin2float(b):
    normalized_hex = b[2:].zfill(8)
    return struct.unpack('>f', bytes.fromhex(normalized_hex))[0]


# -----------------------------------------------------------------------------

def parse_args():
    parser = argparse.ArgumentParser(description="MJK Module Tester/Calibrator")

    default_uart_port = ""
    if platform.system() == 'Darwin':
        default_uart_port = "/dev/tty.usbmodemFD502"
    elif platform.system() == 'Linux':
        default_uart_port = "/dev/ttyACM0"

    parser.add_argument('--port', dest='port', action='store', type=str, default=default_uart_port, required=False,
                        help='MCU UART port (default %s)' % default_uart_port)
    parser.add_argument('--baudrate', dest='baudRate', type=int, default=9600, help='MCU UART baud rate')
    if isVoltmeterSupported:
        parser.add_argument('--voltmeter-port', dest='voltmeter_port', action='store', type=str, default='ASRL4::INSTR',
                            required=False, help='Voltmeter VISA port (default ASRL4::INSTR)')

    parser.add_argument('--verbose', dest='verbose', action='store_true', default=False, required=False,
                        help='Verbose debug logging')
    parser.add_argument('--yes', dest='yes', action='store_true', default=False, required=False,
                        help='Ask no questions')
    parser.add_argument('--samples', dest='num_samples', type=int, default=10, required=False,
                        help='Number of voltmeter samples for each DAC code')
    parser.add_argument('--steps', dest='num_steps', type=int, default=10, required=False,
                        help='Number of voltage steps within the window [min, max] (default 10)')
    parser.add_argument('--min', dest='min', type=float, default=-2.1, required=False,
                        help='Minimum voltage (default -2.1)')
    parser.add_argument('--max', dest='max', type=float, default=+4.1, required=False,
                        help='Maximum voltage (default +4.1)')
    parser.add_argument('--no-plots', dest='render_plots', action='store_false', default=True, required=False,
                        help='Turn off plotting with matplotlib')
    parser.add_argument('mode', choices=[
        "info",
        "inputs",
        "leds",
        "dac",
        "adc",
        "dev",
        "load",
        "save",
        "lut-dac"
    ], help='Mode')
    parser.add_argument('path', metavar='PATH', type=str, nargs='?',
                        help='Output path (JSON for calibration data dump or CSV for LUTs)')

    args = parser.parse_args()

    if args.min >= args.max:
        print_error("min (%s) must be smaller than max (%s)" % (args.min, args.max))
        return False

    if args.num_steps < 3:
        print_error("We need at least 3 steps")
        return False

    return args


# -----------------------------------------------------------------------------
# MARK: -

class Module:
    def __init__(self):
        self.protocol_version = None
        self.manufacturer = None
        self.name = None
        self.hardware_version = None
        self.firmware_version = None
        self.potentiometers = []
        self.cv_inputs = []
        self.gate_inputs = []
        self.gate_normalization = []
        self.switches = []
        self.dac_outputs = []
        self.leds = []

    def fetch_info(self, tester):
        self.protocol_version = tester.get_list(ListSelector.LIST_PROTOCOL_VERSION)
        self.manufacturer = tester.get_list(ListSelector.LIST_MANUFACTURER)[0]
        self.name = tester.get_list(ListSelector.LIST_PRODUCT_NAME)[0]
        self.hardware_version = tester.get_list(ListSelector.LIST_PRODUCT_HARDWARE_VERSION)
        self.firmware_version = tester.get_list(ListSelector.LIST_PRODUCT_FIRMWARE_VERSION)
        self.potentiometers = tester.get_list(ListSelector.LIST_POT)
        self.cv_inputs = tester.get_list(ListSelector.LIST_CV_ADC_CHANNEL)
        self.gate_inputs = tester.get_list(ListSelector.LIST_GATE)
        self.gate_normalization = tester.get_list(ListSelector.LIST_GATE_NORMALIZATION)
        self.switches = tester.get_list(ListSelector.LIST_SWITCH)
        self.dac_outputs = tester.get_list(ListSelector.LIST_DAC_CHANNEL)
        self.leds = tester.get_list(ListSelector.LIST_LED)

    def info_dict(self):
        return self.__dict__

    def json(self):
        return json.dumps(self.info_dict(), indent=4)

    def info(self):
        return '''
\tProtocol version: %s
\tManufacturer: %s
\tName: %s
\tHardware version: %s
\tFirmware version: %s
\tPotentiometers: %s
\tGate: %s
\tGate Normalization: %s
\tSwitches: %s
\tADC CV Channels: %s
\tDAC Channels: %s
\tLEDs: %s
''' % (
            self.protocol_version, self.manufacturer, self.name,
            self.hardware_version, self.firmware_version, self.potentiometers,
            self.gate_inputs, self.gate_normalization, self.switches,
            self.cv_inputs, self.dac_outputs, self.leds)


class Tides(Module):
    def __init__(self):
        super(Tides, self).__init__()
        self.protocol_version = ['0', '1']
        self.manufacturer = "martinjankoehler.com"
        self.name = "MJK Tidal Modulator"
        self.hardware_version = ["v1c"]
        self.firmware_version = ['1', '1']
        self.potentiometers = ["FREQUENCY", "-", "SHAPE", "SLOPE", "SMOOTHNESS",
                               "SHIFT/LEVEL", "-", "FREQUENCY ATT", "SHAPE ATT",
                               "SLOPE ATT", "SMOOTHNESS ATT", "SHIFT/LEVEL ATT"]
        self.cv_inputs = ["V/OCT", "FM", "SHAPE", "SLOPE", "SMOOTHNESS", "SHIFT"]
        self.gate_inputs = ["TRIG", "CLOCK"]
        self.gate_normalization = self.gate_inputs
        self.dac_outputs = ["OUT1", "OUT2", "OUT3", "OUT4"]
        self.switches = ["RANGE", "MODE", "SHIFT"]
        self.leds = ['1', '2', '3', '4', '5', '6', '7', '8', '9']


# -----------------------------------------------------------------------------
# MARK: -

class RequestType(IntEnum):
    CONTROL = 0
    GET_BYTE = 1
    SET_BYTE = 2
    READ_NIBBLE = 3
    WRITE_NIBBLE = 4
    FLUSH_TO_CHANNEL = 5
    ENUMERATE_LIST = 6


# 7: unused


class ControlAction(IntEnum):
    RESET_READ_STATE = 0
    RESET_WRITE_STATE = 1
    RESET_LIST_STATE = 2
    # 3-5: unused
    RESTORE_DEFAULT_SETTINGS = 6  # in RAM (no sanitization)
    LOAD_SETTINGS_FROM_FLASH = 7  # Load to RAM (no sanitization)
    SANITIZE_SETTINGS = 8  # in RAM
    FLASH_SETTINGS = 9
    GENERATE_TEST_SIGNALS = 10
    # 11-16: unused
    ADC_CALIBRATION_STEP1 = 17
    ADC_CALIBRATION_STEP2 = 18
    ADC_CALIBRATION_STEP3 = 19
    LEDS_OFF = 20
    LEDS_ON_ALL_DEFAULT = 21
    LEDS_ON_ALL_RED = 22
    LEDS_ON_ALL_GREEN = 23
    LEDS_ON_ALL_YELLOW = 24


# 25-31: unused


class ByteSource(IntEnum):
    ERROR_CODE = 0
    READ_SELECTOR_VALUE = 1
    WRITE_SELECTOR_VALUE = 2
    LIST_SELECTOR_VALUE = 3
    # 4..7: unused
    NEXT_LIST_EXTRY_BYTE = 8
    LIST_LENGTH = 9
    NUMBER_FORMAT = 10
    # 11...19: unused
    NUMBER_OF_POTS = 20
    NUMBER_OF_GATES = 21
    NUMBER_OF_CV_ADC_CHANNELS = 22
    NUMBER_OF_DAC_CHANNELS = 23


# 24...31: unused


class ByteDestination(IntEnum):
    # RESERVED = 0
    READ_SELECTOR_VALUE = 1
    WRITE_SELECTOR_VALUE = 2
    LIST_SELECTOR_VALUE = 3
    # 4...9: unused
    NUMBER_FORMAT = 10


# 11...31: unused


class ReadSelector(IntEnum):
    READ_SELECTOR_UNINITIALIZED = 0
    # 1, 2: reserved
    READ_POT = 3
    READ_CV = 4
    READ_CV_CALIBRATED = 5
    READ_GATE_NORMALIZATION = 6
    READ_GATE_VALUE = 7
    READ_SWITCH = 8
    READ_CALIBRATION_ADC_SCALE = 9
    READ_CALIBRATION_ADC_OFFSET = 10
    READ_CALIBRATION_DAC_SCALE = 11
    READ_CALIBRATION_DAC_OFFSET = 12
    READ_FORCED_DAC_CODE = 13
    READ_FORCED_DAC_FLOAT_UNCALIBRATED = 14
    READ_FORCED_DAC_FLOAT_CALIBRATED = 15
    # 16...26: unused
    READ_WRITE_REGISTER = 27
    READ_CAFEBABE = 28
    READ_MIN = 29
    READ_ZERO = 30
    READ_MAX = 31


class WriteSelector(IntEnum):
    WRITE_SELECTOR_UNINITIALIZED = 0
    # 1...7: unused
    WRITE_CALIBRATION_ADC_SCALE = 8
    WRITE_CALIBRATION_ADC_OFFSET = 9
    WRITE_CALIBRATION_DAC_SCALE = 10
    WRITE_CALIBRATION_DAC_OFFSET = 11
    WRITE_FORCED_DAC_CODE = 12
    WRITE_FORCED_DAC_FLOAT_UNCALIBRATED = 13
    WRITE_FORCED_DAC_FLOAT_CALIBRATED = 14


# 15...31: unused


class ListSelector(IntEnum):
    LIST_SELECTOR_UNINITIALIZED = 0
    LIST_PROTOCOL_VERSION = 1
    LIST_MANUFACTURER = 2
    LIST_PRODUCT_NAME = 3
    LIST_PRODUCT_HARDWARE_VERSION = 4
    LIST_PRODUCT_FIRMWARE_VERSION = 5
    LIST_POT = 6
    LIST_GATE = 7
    LIST_GATE_NORMALIZATION = 8
    LIST_SWITCH = 9
    LIST_CV_ADC_CHANNEL = 10
    LIST_DAC_CHANNEL = 11
    LIST_LED = 12


# 13...15: unused


class NumberFormat(IntEnum):
    NUMBER_FORMAT_UNINITIALIZED = 0
    BYTE = 1
    UINT16 = 2
    FLOAT = 3
    UINT32 = 4

    # 5-7: unused

    def num_bytes(self):
        if self == NumberFormat.NUMBER_FORMAT_UNINITIALIZED:
            raise Exception("NumberFormat is uninitialized!")
        elif self == NumberFormat.BYTE:
            return 1
        elif self == NumberFormat.UINT16:
            return 2
        elif self == NumberFormat.FLOAT:
            return 4
        elif self == NumberFormat.UINT32:
            return 4


# -----------------------------------------------------------------------------
# MARK: -

class GetByteRequest(Structure):
    _pack_ = 1
    _fields_ = [
        ("source", c_ubyte, 5),
        ("type", c_ubyte, 3),
    ]


class SetByteRequest(Structure):
    _pack_ = 1
    _fields_ = [
        ("selector", c_ubyte, 5),
        ("type", c_ubyte, 3),
    ]


class ControlRequest(Structure):
    _pack_ = 1
    _fields_ = [
        ("action", c_uint8, 5),
        ("type", c_uint8, 3),
    ]


class ReadNibbleRequest(Structure):
    _pack_ = 1
    _fields_ = [
        ("channel", c_uint8, 5),
        ("type", c_uint8, 3),
    ]


class WriteNibbleRequest(Structure):
    _pack_ = 1
    _fields_ = [
        ("nibble", c_uint8, 4),
        ("is_finished", c_uint8, 1),
        ("type", c_uint8, 3),
    ]


class FlushToChannelRequest(Structure):
    _pack_ = 1
    _fields_ = [
        ("channel", c_uint8, 5),
        ("type", c_uint8, 3),
    ]


class EnumerateListRequest(Structure):
    _pack_ = 1
    _fields_ = [
        ("index", c_uint8, 5),
        ("type", c_uint8, 3),
    ]


class ReadNibbleResponse(Structure):
    _pack_ = 1
    _fields_ = [
        ("nibble", c_uint8, 4),
        ("is_finished", c_uint8, 1),
        ("unused", c_uint8, 2),
        ("error_available", c_uint8, 1),
    ]


class EnumerateListResponse(Structure):
    _pack_ = 1
    _fields_ = [
        ("list_length", c_uint8, 5),
        ("unused", c_uint8, 2),
        ("error_available", c_uint8, 1),
    ]


class GenericRequest(Union):
    _pack_ = 1
    _fields_ = [
        ("bytes", c_uint8, 8),
        ("get_byte_request", GetByteRequest),
        ("set_byte_request", SetByteRequest),
        ("control_request", ControlRequest),
        ("read_nibble_request", ReadNibbleRequest),
        ("write_nibble_request", WriteNibbleRequest),
        ("flush_to_channel_request", FlushToChannelRequest),
        ("enumerate_list_request", EnumerateListRequest),
    ]


class GenericResponse(Union):
    _pack_ = 1
    _fields_ = [
        ("bytes", c_uint8, 8),
        ("read_nibble_response", ReadNibbleResponse),
        ("enumerate_list_response", EnumerateListResponse),
    ]


# -----------------------------------------------------------------------------
# MARK: -

class Tester:
    def __init__(self, module, uart, verbose):
        self.module = module
        self.uart = uart
        self.verbose = verbose
        self.last_forced_calibrated_voltage = []
        for channel in range(0, len(self.module.dac_outputs)):
            self.last_forced_calibrated_voltage.append(None)

    def print_dbg(self, s):
        if self.verbose:
            print(s)

    def print_roundtrip(self, topic, input_parameter, request_bytes, output, response_bytes):
        self.print_dbg("%s(%s) [request %s] -> %s [response %s]" %
                       (topic, input_parameter, '{:08b}'.format(request_bytes[0]), output, response_bytes))
        return

    def read(self, num_bytes, topic=None):
        if topic is None:
            topic = "%s()" % inspect.stack()[1][3]

        read_bytes = self.uart.read(num_bytes)
        if len(read_bytes) == num_bytes:
            return read_bytes
        else:
            raise Exception("%s timed out trying to receive %d byte, only got %d" % (topic, num_bytes, len(read_bytes)))

    def get_byte(self, source):
        request = GetByteRequest()
        request.type = RequestType.GET_BYTE
        request.source = source

        packed = GenericRequest()
        packed.get_byte_request = request
        request_bytes = bytes([packed.bytes])

        self.uart.write(request_bytes)
        response_bytes = self.read(1)
        value = int.from_bytes(response_bytes, byteorder='little')

        self.print_roundtrip('get_byte', source, request_bytes, value, response_bytes)
        return value

    def get_read_selector(self):
        v = self.get_byte(ByteSource.READ_SELECTOR_VALUE)
        return ReadSelector(v)

    def set_byte(self, selector, payload):
        request = SetByteRequest()
        request.type = RequestType.SET_BYTE
        request.selector = selector

        packed = GenericRequest()
        packed.set_byte_request = request
        request_bytes = bytes([packed.bytes, payload])

        self.uart.write(request_bytes)
        response_bytes = self.read(len(request_bytes))
        value = int.from_bytes(response_bytes, byteorder='little')
        self.print_roundtrip('SetByte[%s]' % selector, payload, request_bytes, value, response_bytes)
        return value

    def set_read_selector(self, selector):
        return self.set_byte(ByteDestination.READ_SELECTOR_VALUE, selector)

    def get_write_selector(self):
        v = self.get_byte(ByteSource.WRITE_SELECTOR_VALUE)
        return WriteSelector(v)

    def set_write_selector(self, selector):
        return self.set_byte(ByteDestination.WRITE_SELECTOR_VALUE, selector)

    def get_list_selector(self):
        v = self.get_byte(ByteSource.LIST_SELECTOR_VALUE)
        return WriteSelector(v)

    def set_list_selector(self, selector):
        return self.set_byte(ByteDestination.LIST_SELECTOR_VALUE, selector)

    def get_number_format(self):
        v = self.get_byte(ByteSource.NUMBER_FORMAT)
        return NumberFormat(v)

    def set_number_format(self, number_format):
        return self.set_byte(ByteDestination.NUMBER_FORMAT, number_format)

    def control(self, action):
        request = ControlRequest()
        request.type = RequestType.CONTROL
        request.action = action

        packed = GenericRequest()
        packed.control_request = request
        request_bytes = bytes([packed.bytes])

        self.uart.write(request_bytes)
        response_bytes = self.read(1)
        value = int.from_bytes(response_bytes, byteorder='little')
        self.print_roundtrip('Control', action, request_bytes, value, response_bytes)
        return value

    def read_channel_value(self, number_format, channel):
        request_bytes = []

        num_bytes = number_format.num_bytes()
        num_nibbles = num_bytes * 2

        for i in range(0, num_nibbles):
            request = ReadNibbleRequest()
            request.type = RequestType.READ_NIBBLE
            request.channel = channel
            packed = GenericRequest()
            packed.read_nibble_request = request
            request_bytes.append(packed.bytes)
            self.print_dbg('request {:d} [{:08b}]'.format(packed.bytes, packed.bytes))

        self.print_dbg(bytes(request_bytes))
        self.uart.write(bytes(request_bytes))

        response_bytes = self.read(num_nibbles)
        self.print_dbg("response_bytes: %s" % response_bytes.hex())
        nibbles = []
        for i, b in enumerate(response_bytes):
            response = GenericResponse()
            response.bytes = b
            read_nibble_response = response.read_nibble_response
            if read_nibble_response.error_available == 1:
                raise Exception("MCU reported error")

            nibbles.append(read_nibble_response.nibble)
            if i == num_nibbles and not read_nibble_response.is_finished:
                raise Exception("MCU error: expected is_finished")

        payload_bytes = []
        next_byte = 0
        for i, nibble in enumerate(nibbles):
            if i % 2 == 0:
                next_byte = nibble
            else:
                next_byte = (nibble << 4) | next_byte
                payload_bytes.append(next_byte)

        value = None
        if number_format == NumberFormat.FLOAT:
            b = bytes(payload_bytes)
            self.print_dbg("read_channel_value: payload is %s %s" % (b, b.hex()))
            value = struct.unpack('<f', b)[0]
        else:
            self.print_dbg("read_channel_value: payload is %s %s" % (bytes(payload_bytes), bytes(payload_bytes).hex()))
            value = int.from_bytes(payload_bytes, byteorder='little')
        self.print_roundtrip('read_channel_value[%d]' % channel, number_format, request_bytes, value, response_bytes)
        return value

    def write_data(self, number_format, value):
        num_bytes = number_format.num_bytes()
        num_nibbles = num_bytes * 2

        nibbles = []
        if number_format == NumberFormat.FLOAT:
            binary = struct.pack('<f', value)
            for b in binary:
                nibbles.append(b & 0xF)
                nibbles.append((b >> 4) & 0xF)
        else:
            for i in range(0, num_nibbles):
                n = (value >> (i * 4)) & 0xF
                nibbles.append(n)

        self.print_dbg("nibbles: %s" % nibbles)

        request_bytes = []
        for (i, nibble) in enumerate(nibbles):
            request = WriteNibbleRequest()
            request.type = RequestType.WRITE_NIBBLE
            request.nibble = nibble
            request.is_finished = 1 if i + 1 == num_nibbles else 0
            packed = GenericRequest()
            packed.write_nibble_request = request
            request_bytes.append(packed.bytes)
            self.print_dbg('request {:d} [{:08b}]'.format(packed.bytes, packed.bytes))

        # print(bytes(request_bytes))
        self.uart.write(bytes(request_bytes))
        response_bytes = self.uart.read(num_nibbles)
        self.print_roundtrip('wrote_data[%s, %s]' % (number_format, value), number_format, request_bytes, value,
                             response_bytes)

    def flush_data_to_channel(self, channel):
        request = FlushToChannelRequest()
        request.type = RequestType.FLUSH_TO_CHANNEL
        request.channel = channel

        packed = GenericRequest()
        packed.flush_to_channel_request = request
        request_bytes = bytes([packed.bytes])

        self.uart.write(request_bytes)
        response_bytes = self.read(1)
        value = int.from_bytes(response_bytes, byteorder='little')
        self.print_roundtrip('flush_data_to_channel', channel, request_bytes, value, response_bytes)
        return value

    def write_channel_value(self, number_format, channel, value):
        self.write_data(number_format, value)
        self.flush_data_to_channel(channel)

    def get_list_length(self):
        return self.get_byte(ByteSource.LIST_LENGTH)

    def enumerate_list_entry(self, index):
        request = EnumerateListRequest()
        request.type = RequestType.ENUMERATE_LIST
        request.index = index

        packed = GenericRequest()
        packed.enumerate_list_request = request
        request_bytes = bytes([packed.bytes])

        self.uart.write(request_bytes)
        response_bytes = self.read(1)
        value = int.from_bytes(response_bytes, byteorder='little')
        self.print_roundtrip('EnumerateListRequest', index, request_bytes, value, response_bytes)
        return value

    def get_list_entry_next_byte(self):
        return self.get_byte(ByteSource.NEXT_LIST_EXTRY_BYTE)

    def get_list(self, selector):
        entries = []
        self.set_list_selector(selector)
        length = self.get_list_length()
        for i in range(0, length):
            self.enumerate_list_entry(i)
            payload_bytes = []
            b = None
            while b != 0:
                b = self.get_list_entry_next_byte()
                payload_bytes.append(b)
            entries.append(str(bytes(payload_bytes[:-1]), 'utf8'))
        return entries

    # MARK: -
    # ----------------------------------------
    def test_inputs(self):
        self.set_number_format(NumberFormat.UINT16)

        self.dump_values("Potentiometer", ReadSelector.READ_POT, NumberFormat.UINT16, self.module.potentiometers)
        print("-----------------------------------------------------")
        self.dump_values("CV", ReadSelector.READ_CV, NumberFormat.UINT16, self.module.cv_inputs)
        print("-----------------------------------------------------")

        self.dump_gate_input_info()
        print("-----------------------------------------------------")
        self.dump_values("Switches", ReadSelector.READ_SWITCH, NumberFormat.BYTE, self.module.switches)

    def dump_values(self, topic, read_selector, number_format, channels):
        longest_name = max(channels, key=len)
        width = len(longest_name) + 1

        self.set_read_selector(read_selector)

        for channel, name in enumerate(channels):
            if name == "-":
                continue
            value = self.read_channel_value(number_format, channel)
            print("%s %02d %s: %s" % (topic, channel, name.rjust(width, ' '), value))

    def dump_gate_input_info(self):
        number_format = NumberFormat.BYTE
        self.set_number_format(number_format)

        channels = self.module.gate_inputs
        longest_name = max(channels, key=len)
        width = len(longest_name) + 1

        normalization = []

        self.set_read_selector(ReadSelector.READ_GATE_NORMALIZATION)
        for channel, name in enumerate(channels):
            if name == "-":
                continue
            value = self.read_channel_value(number_format, channel)
            value = "connected" if value == 0 else "not connected"
            normalization.append(value)
            print("Gate Normalization %02d %s: %s" % (channel, name.rjust(width, ' '), value))
        print("-----------------------------------------------------")

        self.set_read_selector(ReadSelector.READ_GATE_VALUE)
        for channel, name in enumerate(channels):
            if name == "-":
                continue
            if normalization[channel] == "connected":
                value = self.read_channel_value(number_format, channel)
                value = "on" if value == 1 else "off"
            else:
                value = "off"
            print("Gate Value %02d %s: %s" % (channel, name.rjust(width, ' '), value))

    def force_dac_code(self, code, channel):
        self.set_number_format(NumberFormat.UINT16)
        self.set_write_selector(WriteSelector.WRITE_FORCED_DAC_CODE)
        self.write_channel_value(NumberFormat.UINT16, channel, code)
        self.last_forced_calibrated_voltage[channel] = None

    def force_dac_calibrated(self, value, channel):
        self.set_number_format(NumberFormat.FLOAT)
        self.set_write_selector(WriteSelector.WRITE_FORCED_DAC_FLOAT_CALIBRATED)
        self.write_channel_value(NumberFormat.FLOAT, channel, value)
        self.last_forced_calibrated_voltage[channel] = value

    def force_dac_uncalibrated(self, value, channel):
        self.set_number_format(NumberFormat.UINT16)
        self.set_write_selector(WriteSelector.WRITE_FORCED_DAC_FLOAT_UNCALIBRATED)
        self.write_channel_value(NumberFormat.UINT16, channel, value)
        self.last_forced_calibrated_voltage[channel] = None

    @staticmethod
    def split_calibration_data(value):
        offset = value & 0xFFFF
        scale = (value >> 16) & 0xFFFF
        return [offset, scale]

    @staticmethod
    def prepare_calibration_args(number_format_args, channel_args, all_channels):
        channel_names = None
        if channel_args is None:
            channel_names = all_channels
        elif isinstance(channel_args, int):
            channel_names = [all_channels[channel_args]]
        elif isinstance(channel_args, list):
            if len(channel_args) == 0:
                channel_names = all_channels
            elif isinstance(channel_args[0], int):
                print(channel_args)
                channel_names = list(map(lambda c: all_channels[c], channel_args))
            elif isinstance(channel_args[0], str):
                channel_names = channel_args
        if channel_names is None:
            raise Exception("unsupported channel arguments: %s" % repr(channel_args))

        needed_formats = []
        if isinstance(number_format_args, NumberFormat):
            needed_formats = [number_format_args]
        elif isinstance(number_format_args, list):
            needed_formats = number_format_args
        return channel_names, needed_formats

    def calibration_data(self, number_format_args, scale_selector, offset_selector, channel_args, all_channels):
        channel_names, needed_formats = self.prepare_calibration_args(number_format_args, channel_args, all_channels)
        format_description = {NumberFormat.UINT32: 'bin', NumberFormat.FLOAT: 'float'}

        data = {}
        for name in channel_names:
            data[name] = {'offset': {}, 'scale': {}}

        for f in needed_formats:
            self.set_number_format(f)

            self.set_read_selector(scale_selector)
            for name in channel_names:
                channel = all_channels.index(name)
                v = self.read_channel_value(f, channel)
                if f == NumberFormat.UINT32:
                    v = "0x%X" % v
                data[name]['scale'][format_description[f]] = v

            self.set_read_selector(offset_selector)
            for name in channel_names:
                channel = all_channels.index(name)
                v = self.read_channel_value(f, channel)
                if f == NumberFormat.UINT32:
                    v = "0x%X" % v
                data[name]['offset'][format_description[f]] = v

        return data

    def dac_calibration_data(self, number_format_args=(NumberFormat.FLOAT, NumberFormat.UINT32), channel_args=None):
        return self.calibration_data(number_format_args, ReadSelector.READ_CALIBRATION_DAC_SCALE,
                                     ReadSelector.READ_CALIBRATION_DAC_OFFSET, channel_args, self.module.dac_outputs)

    def adc_calibration_data(self, number_format_args=(NumberFormat.FLOAT, NumberFormat.UINT32), channel_args=None):
        return self.calibration_data(number_format_args, ReadSelector.READ_CALIBRATION_ADC_SCALE,
                                     ReadSelector.READ_CALIBRATION_ADC_OFFSET, channel_args, self.module.cv_inputs)

    def print_dac_calibration_data(self, channel_args=None):
        number_format_args = [NumberFormat.FLOAT, NumberFormat.UINT32]
        d = self.dac_calibration_data(number_format_args, channel_args)
        channel_names, needed_formats = self.prepare_calibration_args(number_format_args, channel_args,
                                                                      self.module.dac_outputs)

        longest_name = max(channel_names, key=len)
        width = len(longest_name) + 1

        print("DAC calibration data:")
        for i, name in enumerate(channel_names):
            channel_number = self.module.dac_outputs.index(name)
            s = d[name]['scale']
            o = d[name]['offset']
            print("  DAC[%d] %4s, scale: %s [%s], offset: %s [%s]" %
                  (channel_number, name.rjust(width, ' '), s['float'], s['bin'], o['float'], o['bin']))

    def print_adc_calibration_data(self, channel_args=None):
        number_format_args = [NumberFormat.FLOAT, NumberFormat.UINT32]
        d = self.adc_calibration_data(number_format_args, channel_args)
        channel_names, needed_formats = self.prepare_calibration_args(number_format_args, channel_args,
                                                                      self.module.cv_inputs)

        longest_name = max(channel_names, key=len)
        width = len(longest_name) + 1

        print("ADC calibration data:")
        for i, name in enumerate(channel_names):
            channel_number = self.module.cv_inputs.index(name)
            s = d[name]['scale']
            o = d[name]['offset']
            print("  ADC[%d] %4s, scale: %s [%s], offset: %s [%s]" %
                  (channel_number, name.rjust(width, ' '), s['float'], s['bin'], o['float'], o['bin']))

    def calibration_data_dump(self, number_format):
        d = {
            'dac': self.dac_calibration_data(number_format, self.module.dac_outputs),
            'adc': self.adc_calibration_data(number_format, self.module.cv_inputs)
        }
        return d

    def print_calibration_data_json(self, number_format=(NumberFormat.FLOAT, NumberFormat.UINT32)):
        d = self.calibration_data_dump(number_format)
        json_data = json.dumps(d, indent=4)
        print(json_data)

    def validate_calibration_data_dump(self, d):
        if "dac" not in d:
            return False
        if "adc" not in d:
            return False
        for name in self.module.dac_outputs:
            if name not in d["dac"]:
                return False
            if "offset" not in d["dac"][name]:
                return False
            if "scale" not in d["dac"][name]:
                return False
            if "bin" not in d["dac"][name]["offset"]:
                return False
            if "bin" not in d["dac"][name]["scale"]:
                return False
        for name in self.module.cv_inputs:
            if name not in d["adc"]:
                return False
            if "offset" not in d["adc"][name]:
                return False
            if "scale" not in d["adc"][name]:
                return False
            if "bin" not in d["adc"][name]["offset"]:
                return False
            if "bin" not in d["adc"][name]["scale"]:
                return False
        return True

    def restore_calibration_data(self, dump):
        if not self.validate_calibration_data_dump(dump):
            print_error("Failed to validate calibration dump")
            return False

        self.set_number_format(NumberFormat.UINT32)
        self.set_write_selector(WriteSelector.WRITE_CALIBRATION_DAC_SCALE)
        for (channel, name) in enumerate(self.module.dac_outputs):
            value = int(dump["dac"][name]["scale"]["bin"], base=16)
            self.write_channel_value(NumberFormat.UINT32, channel, value)

        self.set_write_selector(WriteSelector.WRITE_CALIBRATION_DAC_OFFSET)
        for (channel, name) in enumerate(self.module.dac_outputs):
            value = int(dump["dac"][name]["offset"]["bin"], base=16)
            self.write_channel_value(NumberFormat.UINT32, channel, value)

        self.set_write_selector(WriteSelector.WRITE_CALIBRATION_ADC_SCALE)
        for (channel, name) in enumerate(self.module.cv_inputs):
            value = int(dump["adc"][name]["scale"]["bin"], base=16)
            self.write_channel_value(NumberFormat.UINT32, channel, value)

        self.set_write_selector(WriteSelector.WRITE_CALIBRATION_ADC_OFFSET)
        for (channel, name) in enumerate(self.module.cv_inputs):
            value = int(dump["adc"][name]["offset"]["bin"], base=16)
            self.write_channel_value(NumberFormat.UINT32, channel, value)

        self.control(ControlAction.FLASH_SETTINGS)

        return True


# -----------------------------------------------------------------------------
# MARK: -

def info_mode(tester):
    connected_module = Module()
    connected_module.fetch_info(tester)

    a = connected_module.info()
    b = tester.module.info()
    if a == b:
        print("MATCH: Connected module matches expectation!\n")
    else:
        print_error("Connected module does not match expectation:")
        print("%s" % term_color_yellow, end='')
        for line in difflib.unified_diff(a.split("\n"), b.split("\n")):
            print(line)
        print("%s" % term_color_reset, end='')

    print("-----------------------------------------------------")
    print("Connected Module:\n%s" % a)

    print("-----------------------------------------------------")
    print("Expected Module (cached):\n%s" % b)


def inputs_mode(args, tester):
    while True:
        os.system('clear')
        print("%s - Test Mode '%s'" % (tester.module.name, args.mode))
        print("-----------------------------------------------------")

        # set inputs so that we can test different CV values
        tester.force_dac_code(0xFFFF, 0)  # CH0 @ -8.16V
        tester.force_dac_calibrated(-2.0, 1)  # CH1 @ -2V
        tester.force_dac_calibrated(0.0, 2)  # CH2 @ 0V
        tester.force_dac_code(0, 3)  # CH3 @ circa +8.16V

        tester.test_inputs()
        time.sleep(1)


def led_mode(args, tester):
    os.system('clear')
    print("%s - Test Mode '%s'" % (tester.module.name, args.mode))
    print("-----------------------------------------------------")
    print("NOTE: Type q⮐  to quit")

    action_regex = None

    while True:
        print("-----------------------------------------------------")
        if action_regex is None:
            action_regex = re.compile(r'([0orgydq])')

        print("\n" +
              "\tAvailable control actions:\n" +
              "\t\to) LEDs off\n" +
              "\t\tr) LEDs red\n" +
              "\t\tg) LEDs green\n" +
              "\t\ty) LEDs yellow\n" +
              "\t\td) LEDs default colors\n" +
              "\t\t---\n" +
              "\t\tq) Quit menu\n")

        action = input(prompt("LED action"))

        match = action_regex.fullmatch(action)
        if match is None:
            print_error("Invalid action '%s'" % action)
            continue

        action = match.group(1)

        if action == "q":
            break
        if action == "0" or action == "o":
            tester.control(ControlAction.LEDS_OFF)
        elif action == "r":
            tester.control(ControlAction.LEDS_ON_ALL_RED)
        elif action == "g":
            tester.control(ControlAction.LEDS_ON_ALL_GREEN)
        elif action == "y":
            tester.control(ControlAction.LEDS_ON_ALL_YELLOW)
        elif action == "d":
            tester.control(ControlAction.LEDS_ON_ALL_DEFAULT)


# -----------------------------------------------------------------------------
# MARK: -

class CalibrationParameter(IntEnum):
    SCALE = 0
    OFFSET = 1


class VoltageMeasurement:
    def __init__(self, parameter_codes, minus2=-2.0, zero=0.0, plus4=+4.0):
        self.code = parameter_codes

        self.minus2 = minus2
        self.zero = zero
        self.plus4 = plus4

    def error(self, parameter):
        if parameter == CalibrationParameter.SCALE:
            return (self.plus4 - 4.0) + (self.minus2 - -2.0)
        elif parameter == CalibrationParameter.OFFSET:
            return self.zero
        else:
            raise Exception("Invalid calibration parameter:", parameter)

    def improvement(self, old_measurement, parameter):
        SCALE = CalibrationParameter.SCALE
        OFFSET = CalibrationParameter.OFFSET

        a = abs(self.error(parameter))
        b = abs(old_measurement.error(parameter))
        if a == b:
            other_parameter = OFFSET if parameter == SCALE else SCALE
            a = abs(self.error(other_parameter))
            b = abs(old_measurement.error(other_parameter))
            return a < b
        else:
            return a < b

    def print(self, name, old_measurement=None):
        SCALE = CalibrationParameter.SCALE
        OFFSET = CalibrationParameter.OFFSET

        error = {
            'offset_error': self.error(OFFSET),
            'scale_error': self.error(SCALE)
        }
        scale = self.code[SCALE]
        scale_hex = int(struct.pack(">f", scale).hex(), 16)
        offset = self.code[OFFSET]
        offset_hex = int(struct.pack(">f", offset).hex(), 16)
        print("%s scale %s [0x%X], offset %s [0x%X]:" % (name, scale, scale_hex, offset, offset_hex))
        print("\t{ '-2V': %s, '0V': %s, '+4V': %s }" % (self.minus2, self.zero, self.plus4))
        print("\t%s" % error)
        if old_measurement is not None:
            improvement = {
                'offset_improvement': self.improvement(old_measurement, OFFSET),
                'scale_improvement': self.improvement(old_measurement, SCALE),
            }
            print("\t%s" % improvement)


class LinearRegressionResult:
    def __init__(self, measured_voltages, dac_codes):
        scale, offset, r, p, std_err = stats.linregress(measured_voltages, dac_codes)
        self.scale = scale
        self.offset = offset
        self.r = r
        self.p = p
        self.std_err = std_err

    def print(self):
        print(self.__dict__)


class VoltageMeasurements:
    def __init__(self, channel, expected_voltages, scale, offset):
        self.channel = channel
        self.expected_voltages = expected_voltages
        self.scale = scale
        self.offset = offset
        self.dac_codes = []
        self.measured_voltages = []
        self.linear_regression = None

    def dac_code(self, voltage):
        def constrain(val, min_val, max_val):
            return min(max_val, max(min_val, val))

        code = voltage * self.scale + self.offset
        return int(constrain(code, 0, 65535))

    def measure(self, tester, meter, args):
        self.dac_codes.clear()
        self.measured_voltages.clear()
        for v in self.expected_voltages:
            dac_code = self.dac_code(v)
            self.dac_codes.append(dac_code)

            tester.force_dac_code(dac_code, self.channel)
            delay = 0.1
            time.sleep(delay)
            sampled_voltages = meter.voltage(args.num_samples)
            median = statistics.median(sampled_voltages)
            self.measured_voltages.append(median)

        channel_name = tester.module.dac_outputs[self.channel]
        if args.verbose:
            print("DAC[#%d %s]:\n\texpected: %s\n\tdac_codes: %s\n\tmeasured: %s" %
                  (self.channel, channel_name, self.expected_voltages, self.dac_codes, self.measured_voltages))
        self.linear_regression = LinearRegressionResult(self.measured_voltages, self.dac_codes)
        return self.linear_regression


# -----------------------------------------------------------------------------
# MARK: -

class Autotuner:
    def __init__(self, tester, dac_channel, meter, args):
        self.tester = tester
        self.dac_channel = dac_channel
        self.voltmeter = meter
        self.args = args

    @staticmethod
    def generate_voltages(from_voltage, to_voltage, num_steps):
        voltages = []
        if num_steps > 1:
            step = (to_voltage - from_voltage) / num_steps
            for i in range(0, num_steps):
                v = from_voltage + step * i
                voltages.append(v)
        voltages.sort()
        return voltages

    def autotune(self, scale, offset) -> VoltageMeasurements:
        expected_voltages = self.generate_voltages(self.args.from_voltage, self.args.to_voltage, self.args.num_steps)

        m = VoltageMeasurements(self.dac_channel, expected_voltages, scale, offset)
        reg = m.measure(self.tester, self.voltmeter, self.args)
        if self.args.verbose:
            reg.print()

        return m

    def plot(self, measurements):
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        ax.grid(linestyle="--", linewidth=0.5, color='.25', zorder=-10)

        channel = self.dac_channel
        channel_name = self.tester.module.dac_outputs[channel]

        ax.set_title("DAC[#%d %s] calibration" % (channel, channel_name), fontsize=20, verticalalignment='bottom')
        ax.set_xlabel('DAC Code [uint16]')
        ax.set_ylabel('Voltage [V]')
        ax.tick_params(which='major', width=1.0)
        ax.tick_params(which='major', length=10)
        ax.tick_params(which='minor', width=1.0, labelsize=10)
        ax.tick_params(which='minor', length=5, labelsize=10, labelcolor='0.25')

        colors = [
            [None, 'red'],
            ['green', 'yellow']
        ]

        for i, m in enumerate(measurements):
            reg = m.linear_regression

            def calibration(v):
                return reg.scale * v + reg.offset

            descr = "  - scale=%s\n  - offset=%s\n  - r=%s\n  - std_err=%s" % \
                    (reg.scale, reg.offset, reg.r, reg.std_err)
            model = list(map(calibration, m.measured_voltages))
            ax.scatter(m.dac_codes, m.measured_voltages, color=colors[i][0], label="Measured Voltages (Median)")
            iteration = "" if len(measurements) == 0 else " (Iteration #%d)" % (i + 1)
            ax.plot(model, m.measured_voltages, color=colors[i][1],
                    label="Linear Regression%s\n%s\n" % (iteration, descr))

        legend = ax.legend(loc='upper right', shadow=True)  # , fontsize='x-large')
        legend.get_frame().set_facecolor('lightgrey')

        plt.show()

    def setup_mcu_calibration(self, scale, offset):
        self.tester.set_number_format(NumberFormat.FLOAT)
        self.tester.set_write_selector(WriteSelector.WRITE_CALIBRATION_DAC_SCALE)
        self.tester.write_channel_value(NumberFormat.FLOAT, self.dac_channel, scale)
        self.tester.set_write_selector(WriteSelector.WRITE_CALIBRATION_DAC_OFFSET)
        self.tester.write_channel_value(NumberFormat.FLOAT, self.dac_channel, offset)

    def verify(self, scale, offset):
        self.setup_mcu_calibration(scale, offset)

        delay = 0.1

        self.tester.force_dac_calibrated(-2.0, self.dac_channel)
        time.sleep(delay)
        measured_minus2v = self.voltmeter.voltage(self.args.num_samples)

        self.tester.force_dac_calibrated(0.0, self.dac_channel)
        time.sleep(delay)
        measured_zero = self.voltmeter.voltage(self.args.num_samples)

        self.tester.force_dac_calibrated(+4.0, self.dac_channel)
        time.sleep(delay)
        measured_plus4v = self.voltmeter.voltage(self.args.num_samples)

        median_minus2v = statistics.median(measured_minus2v)
        median_zero = statistics.median(measured_zero)
        median_plus4v = statistics.median(measured_plus4v)
        codes = {
            CalibrationParameter.SCALE: scale,
            CalibrationParameter.OFFSET: offset
        }
        m = VoltageMeasurement(codes, minus2=median_minus2v, zero=median_zero, plus4=median_plus4v)
        return m

    def start(self):
        # default coefficients:
        scale = -4032.9
        offset = 32768.0

        first_result = self.autotune(scale, offset)
        reg1 = first_result.linear_regression

        second_result = self.autotune(reg1.scale, reg1.offset)
        reg2 = second_result.linear_regression

        calibrated1 = self.verify(reg1.scale, reg1.offset)
        calibrated2 = self.verify(reg2.scale, reg2.offset)

        calibrated1.print("First iteration:")
        calibrated2.print("Second iteration:", calibrated1)

        # NOTE: we tolerate <1mV offset
        if not calibrated2.improvement(calibrated1, CalibrationParameter.SCALE):
            print_attention("Scale regression with 2nd iteration detected!")
            if calibrated1.error(CalibrationParameter.OFFSET) < 1.0:
                print("Offset of 1st iteration was <1mV, so we use the 1st!")
                self.setup_mcu_calibration(reg1.scale, reg1.offset)

        if self.args.render_plots:
            try:
                self.plot([first_result, second_result])
            except ImportError:
                print_error("Unable to plot results: matplotlib module must be installed!")
                print("\t(Hints:\n" +
                      "\t a) Install matplotlib, for example (macOS): pip3 install matplotlib\n" +
                      "\t b) call script with --no-plots")


# -----------------------------------------------------------------------------
# MARK: -

def build_dac_lut(tester, meter, args):
    delay = 0.1

    print("DAC channels: %s" % tester.module.dac_outputs)

    with open(args.path, 'w') as file:
        csv_writer = csv.writer(file)
        csv_writer.writerow(["DAC Channel", "DAC Code", "Measured Voltage (Avg)", "Measured Voltage (Median)",
                             "Measured Voltages (%d samples)" % args.num_samples])
        progress = None
        for channel, name in enumerate(tester.module.dac_outputs):
            print_attention("Please connect the voltmeter to jack %s" % name)
            answer = input(prompt("Proceed (y/n/q): ")).lower()
            if answer != "y":
                return
            print("Profiling channel DAC[%d] %s ..." % (channel, name), end='')
            for code in range(0, 2 ** 16, args.step):
                tester.force_dac_code(code, channel)
                time.sleep(delay)
                measured_voltages = meter.voltage(args.num_samples)
                mean = statistics.mean(measured_voltages)
                median = statistics.median(measured_voltages)
                csv_writer.writerow([channel, code, mean, median, measured_voltages])
                new_progress = int((code / 2 ** 16) * 100)
                if new_progress != progress:
                    print("%d%%. " % new_progress, end='')
                    progress = new_progress
            print("")


# -----------------------------------------------------------------------------
# MARK: -

def prepare_voltmeter(args):
    import pyvisa

    resource_manager = pyvisa.ResourceManager()
    instrument = resource_manager.open_resource(
        args.voltmeter_port,
        read_termination='\r',
        write_termination='\r\n',
        baud_rate=9600
    )

    meter = voltmeter.Keithley2015Multimeter(instrument)
    meter.reset()

    print("Detected voltmeter: %s" % meter.identification())
    print("---------------------------------------------------------------------")
    time.sleep(1)

    return meter


def autotune(args, tester, channel_filter=None, yes_mode=False):
    meter = prepare_voltmeter(args)
    for channel, name in enumerate(tester.module.dac_outputs):
        if channel_filter is not None and channel not in channel_filter:
            continue
        if not yes_mode:
            print_attention("Please connect the voltmeter to jack %s" % name)
            answer = input(prompt("Proceed with Autotune (y/n/q): ")).lower()
            if answer != "y":
                return

        at = Autotuner(tester, channel, meter, args)
        at.start()
        print("---------------------------------------------------------------------")

    meter.close()


def measure(args, tester, channel_filter=None, yes_mode=False):
    meter = prepare_voltmeter(args)

    calibration = tester.dac_calibration_data()

    for channel, name in enumerate(tester.module.dac_outputs):
        if channel_filter is not None and channel not in channel_filter:
            continue
        if not yes_mode:
            print_attention("Please connect the voltmeter to jack %s" % name)
            answer = input(prompt("Start measuring (y/n/q): ")).lower()
            if answer != "y":
                return

        at = Autotuner(tester, channel, meter, args)
        scale = calibration[name]['scale']['float']
        offset = calibration[name]['offset']['float']
        calibrated1 = at.verify(scale, offset)
        calibrated1.print("DAC[#%d %s]:\n\t" % (channel, name))

        print("---------------------------------------------------------------------")

    meter.close()


def dac_mode(args, tester):
    os.system('clear')
    print("%s - Test Mode '%s'" % (tester.module.name, args.mode))
    print("-----------------------------------------------------")
    print("NOTE: Type q⮐  to quit")

    main_action_regex = None
    action_regex = None

    while True:
        print("-----------------------------------------------------")

        tester.print_dac_calibration_data()

        print("-----------------------------------------------------")

        if main_action_regex is None:
            main_action_regex = re.compile(r'([0123qd]|load|defaults|flash|sanitize|auto|measure|dump|quit)')

        print("\n" +
              "\tAvailable calibration actions:\n" +
              "\t\t0) Enter DAC[0] menu\n" +
              "\t\t1) Enter DAC[1] menu\n" +
              "\t\t2) Enter DAC[2] menu\n" +
              "\t\t3) Enter DAC[3] menu\n" +
              "\t\t---")

        if isVoltmeterSupported:
            print("\t\tauto) Autotune using external voltmeter\n" +
                  "\t\tmeasure) Measure voltages for -2V, 0V, +4V\n" +
                  "\t\t---")

        print("\t\tload) Load settings from flash to RAM\n" +
              "\t\tsanitize) Sanitize settings in RAM\n" +
              "\t\tdefaults) Restore default settings in RAM\n" +
              "\t\tflash) Flash settings\n" +
              "\t\t---\n" +
              "\t\td[ump]) Dump ADC/DAC calibration data\n" +
              "\t\tq[quit]) Quit ADC menu\n")

        action = input(prompt("DAC action"))
        match = main_action_regex.fullmatch(action)
        if match is None:
            print_error("Invalid action '%s'" % action)
            continue

        action = match.group(1)
        if action in ["q", "quit"]:
            sys.exit(1)
        elif action in ["d", "dump"]:
            tester.print_calibration_data_json()
            continue
        elif action == "sanitize":
            tester.control(ControlAction.SANITIZE_SETTINGS)
            continue
        elif action == "defaults":
            tester.control(ControlAction.RESTORE_DEFAULT_SETTINGS)
            continue
        elif action == "load":
            tester.control(ControlAction.LOAD_SETTINGS_FROM_FLASH)
            continue
        elif action == "flash":
            tester.control(ControlAction.FLASH_SETTINGS)
            continue
        elif action == "auto":
            autotune(args, tester)
            continue
        elif action == "measure":
            measure(args, tester)
            continue
        else:
            channel = int(action)

        while True:
            tester.print_dac_calibration_data(channel_args=channel)

            if action_regex is None:
                action_regex = re.compile(
                    r'([01245678dq]|load|defaults|flash|sanitize|auto|measure|dump|quit)|' +
                    r'([c]|code) (\d+)|([osc]|offset|scale]code) (0x[0-9A-Fa-f]+)|' +
                    r'([vso]|voltage|scale|offset) ([+\-]?\d*\.?[0-9e+\-]*)')

            print("\n" +
                  "\tAvailable calibration actions:\n" +
                  "\t\t0) Force calibrated 0V\n" +
                  "\t\t1) Force calibrated -2V\n" +
                  "\t\t2) Force calibrated 4V\n" +
                  "\t\tvoltage <float>) Force calibrated voltage\n" +
                  "\t\t---\n" +
                  "\t\t4) Force -8,1V-ish [DAC code 0xffff]\n" +
                  "\t\t5) Force -2V-ish [DAC code 0x9ff1]\n" +
                  "\t\t6) Force 0V-ish [DAC code 0x7fff]\n" +
                  "\t\t7) Force +4V-ish [DAC code 0x416b]\n" +
                  "\t\t8) Force +8,1V-ish [DAC code 0]\n" +
                  "\t\tcode <n>) Force 16-bit DAC code [example: 0x7fff]\n" +
                  "\t\t---\n" +
                  "\t\toffset <o>) Set offset calibration [0x00000000-0xFFFFFFFF / float]\n" +
                  "\t\tscale <s>)  Set scale calibration  [0x00000000-0xFFFFFFFF / float]")

            if isVoltmeterSupported:
                print("\t\tauto) Autotune using external voltmeter\n" +
                      "\t\tmeasure) Measure voltages for -2V, 0V, +4V\n" +
                      "\t\t---\n")

            print("\t\t---\n" +
                  "\t\tload) Load settings from flash to RAM\n" +
                  "\t\tsanitize) Sanitize settings in RAM\n" +
                  "\t\tdefaults) Restore default settings in RAM\n" +
                  "\t\tflash) Flash settings\n" +
                  "\t\t---\n" +
                  "\t\td[ump]) Dump ADC/DAC calibration data\n" +
                  "\t\tq[uit]) Quit DAC[%d] menu\n" % channel)

            action = input(prompt("DAC[%d] action" % channel))

            match = action_regex.fullmatch(action)
            if match is None:
                print_error("Invalid action '%s'" % action)
                continue

            action = match.group(1)
            value = None
            group3 = match.group(3)
            group5 = match.group(5)
            group7 = match.group(7)
            if group3 is not None:
                action = match.group(2)
                value = int(group3)
            elif group5 is not None:
                action = match.group(4)
                value = int(group5, 16)
            elif group7 is not None:
                action = match.group(6)
                value = float(group7)

            if action in ["q", "quit"]:
                break
            elif action == "0":
                tester.force_dac_calibrated(0.0, channel)
            elif action == "1":
                tester.force_dac_calibrated(-2.0, channel)
            elif action == "2":
                tester.force_dac_calibrated(4.0, channel)
            elif action in ["v", "voltage"]:
                tester.force_dac_calibrated(value, channel)
            elif action == "4":
                tester.force_dac_code(0xffff, channel)
            elif action == "5":
                tester.force_dac_code(0x9ff1, channel)
            elif action == "6":
                tester.force_dac_code(0x7fff, channel)
            elif action == "7":
                tester.force_dac_code(0x416b, channel)
            elif action == "8":
                tester.force_dac_code(0, channel)
            elif action in ["c", "code"]:
                tester.force_dac_code(value, channel)
            elif action == "auto":
                autotune(args, tester, channel_filter=[channel], yes_mode=args.yes)
            elif action == "measure":
                measure(args, tester, channel_filter=[channel], yes_mode=args.yes)
            elif action == "sanitize":
                tester.control(ControlAction.SANITIZE_SETTINGS)
            elif action == "defaults":
                tester.control(ControlAction.RESTORE_DEFAULT_SETTINGS)
            elif action == "load":
                tester.control(ControlAction.LOAD_SETTINGS_FROM_FLASH)
            elif action == "flash":
                tester.control(ControlAction.FLASH_SETTINGS)
            elif action in ["d", "dump"]:
                tester.print_calibration_data_json()
            elif action in ["s", "scale", "o", "offset"]:
                f = None
                if isinstance(value, float):
                    f = NumberFormat.FLOAT
                elif isinstance(value, int):
                    f = NumberFormat.UINT32
                    if 0 <= value and value <= 2 ** 32 - 1:
                        pass
                    else:
                        print_error("number %d is not in [0, %d]" % (value, 2 ** 32 - 1))
                        continue
                else:
                    print_error("unknown number format of value %s" % value)
                    continue

                tester.set_number_format(f)
                if action in ["s", "scale"]:
                    tester.set_write_selector(WriteSelector.WRITE_CALIBRATION_DAC_SCALE)
                    tester.write_channel_value(f, channel, value)
                elif action in ["o", "offset"]:
                    tester.set_write_selector(WriteSelector.WRITE_CALIBRATION_DAC_OFFSET)
                    tester.write_channel_value(f, channel, value)
                last_voltage = tester.last_forced_calibrated_voltage[channel]
                if last_voltage is not None:
                    tester.force_dac_calibrated(last_voltage, channel)

            print("")


def adc_mode(args, tester):
    action_regex = None
    while True:
        print("%s - Test Mode '%s'" % (tester.module.name, args.mode))
        print("-----------------------------------------------------")

        tester.print_adc_calibration_data()

        print("-----------------------------------------------------")

        if action_regex is None:
            action_regex = re.compile(r'([012qd]|all|load|defaults|flash|sanitize|dump|quit)')

        print("\n" +
              "\tAvailable calibration actions:\n" +
              "\t\t---\n" +
              "\t\t0) Step 0:\n" +
              "\t\t\t- Set DAC[0] to -2V\n" +
              "\t\t\t- Disable normalization probe\n" +
              "\t\t1) Step 1:\n" +
              "\t\t\t- Read ADC[0] value for -2V\n" +
              "\t\t\t- Acquire zero offset for other channels\n" +
              "\t\t\t- Persist zero offset for other channels\n" +
              "\t\t\t- Set DAC[0] to +4V\n" +
              "\t\t2) Step 2: \n" +
              "\t\t\t- Read ADC[0] value for +4V\n" +
              "\t\t\t- Autogenerate Calibration Data for ADC[0]\n" +
              "\t\t\t- Persist Calibration Data for ADC[0]\n" +
              "\t\t\t- Restore normalization probe\n" +
              "\t\t---\n" +
              "\t\tall) Run all steps\n" +
              "\t\t---\n" +
              "\t\tload) Load settings from flash to RAM\n" +
              "\t\tsanitize) Sanitize settings in RAM\n" +
              "\t\tdefaults) Restore default settings in RAM\n" +
              "\t\tflash) Flash settings\n" +
              "\t\t---\n" +
              "\t\td[ump]) Dump ADC/DAC calibration data\n" +
              "\t\tq[quit]) Quit ADC menu\n")

        action = input(prompt("ADC action"))

        match = action_regex.fullmatch(action)
        if match is None:
            print_error("Invalid action '%s'" % action)
            continue

        action = match.group(1)

        def adc_calibrate(steps):
            print_attention("Jacks must be connected via patch cable: OUT1 -> V/OCT")
            print("\tPrecondition: DAC[#%d] %s is already calibrated!" % (0, tester.module.dac_outputs[0]))
            if not args.yes:
                answer = input(prompt("Proceed with ADC Calibration? (y/n/q): ")).lower()
                if answer == "y":
                    pass
                elif answer == "q":
                    sys.exit(1)
                else:
                    return
            for step in steps:
                print("\tExecuting %s..." % step)
                tester.control(step)
                time.sleep(1)
            print("\tFinished calibration steps.\n")

        if action in ["q", "quit"]:
            break
        elif action == "all":
            adc_calibrate([
                ControlAction.ADC_CALIBRATION_STEP1,
                ControlAction.ADC_CALIBRATION_STEP2,
                ControlAction.ADC_CALIBRATION_STEP3
            ])
        elif action == "0":
            adc_calibrate([ControlAction.ADC_CALIBRATION_STEP1])
        elif action == "1":
            adc_calibrate([ControlAction.ADC_CALIBRATION_STEP2])
        elif action == "2":
            adc_calibrate([ControlAction.ADC_CALIBRATION_STEP3])
        elif action in ["d", "dump"]:
            tester.print_calibration_data_json()
        elif action == "sanitize":
            tester.control(ControlAction.SANITIZE_SETTINGS)
        elif action == "defaults":
            tester.control(ControlAction.RESTORE_DEFAULT_SETTINGS)
        elif action == "load":
            tester.control(ControlAction.LOAD_SETTINGS_FROM_FLASH)
        elif action == "flash":
            tester.control(ControlAction.FLASH_SETTINGS)


def dev_mode(args, tester):
    action_regex = None
    while True:
        print("%s - Test Mode '%s'" % (tester.module.name, args.mode))
        print("-----------------------------------------------------")

        if action_regex is None:
            action_regex = re.compile(r'^([0-8qd]|1[0-7]|2[0-2]|3[0-5]|4[0-3])$')

        print("\n" +
              "\tAvailable developer tests:\n" +
              "\t\t0) Test get number of DAC channels\n" +
              "\t\t1) Test set/get read selector\n" +
              "\t\t2) Test set/get write selector\n" +
              "\t\t3) Test set/get number format\n" +
              "\t\t---\n" +
              "\t\t4) Test control action LEDS_OFF\n" +
              "\t\t5) Test control action LEDS_ON_ALL_RED\n" +
              "\t\t6) Test control action LEDS_ON_ALL_GREEN\n" +
              "\t\t7) Test control action LEDS_ON_ALL_YELLOW\n" +
              "\t\t8) Test control action LEDS_ON_ALL_DEFAULT\n" +
              "\t\t---\n" +
              "\t\t10) Read UINT8 MAX\n" +
              "\t\t11) Read UINT16 MAX\n" +
              "\t\t12) Read UINT32 MAX\n" +
              "\t\t13) Read FLOAT MAX\n" +
              "\t\t14) Read FLOAT MIN\n" +
              "\t\t15) Read UINT8 0xCAFEBABE\n" +
              "\t\t16) Read UINT16 0xCAFEBABE\n" +
              "\t\t17) Read UINT32 0xCAFEBABE\n" +
              "\t\t---\n" +
              "\t\t20) Read POT[FREQUENCY]\n" +
              "\t\t21) Read ADC[V/OCT]\n" +
              "\t\t22) Read SWITCH[0]\n" +
              "\t\t---\n" +
              "\t\t30) Write/Read 0xCAFEBABE from write buffer\n" +
              "\t\t33) Read float calibration data value of ADC[FM/offset]\n" +
              "\t\t---\n" +
              "\t\t40) Length of manufacturer 'list'\n" +
              "\t\t41) Length of pots 'list'\n" +
              "\t\t42) Read name of manufacturer\n" +
              "\t\t43) Read names of pots\n" +
              "\t\t---\n" +
              "\t\td) Dump ADC/DAC calibration data\n" +
              "\t\t---\n" +
              "\t\tq) Quit ADC menu\n")

        action = input(prompt("Developer action"))

        match = action_regex.fullmatch(action)
        if match is None:
            print_error("Invalid action '%s'" % action)
            continue

        action = match.group(1)

        if action == "q":
            break
        if action == "0":
            n = tester.get_byte(ByteSource.NUMBER_OF_DAC_CHANNELS)
            print("\tGet number of DAC channels: %d" % n)
        elif action == "1":
            s = tester.get_read_selector()
            print("\tGet read selector: %s [%d]" % (s, s))
            for v in [ReadSelector.READ_POT, ReadSelector.READ_CV, ReadSelector.READ_FORCED_DAC_FLOAT_CALIBRATED]:
                print("\tSet read selector to %s [%d]" % (v, v))
                tester.set_read_selector(v)
                s = tester.get_read_selector()
                print("\tGet read selector: %s [%d]" % (s, s))
        elif action == "2":
            s = tester.get_write_selector()
            print("\tGet write selector: %s [%d]" % (s, s))
            for v in [WriteSelector.WRITE_CALIBRATION_ADC_SCALE, WriteSelector.WRITE_FORCED_DAC_FLOAT_CALIBRATED]:
                print("\tSet write selector to %s [%d]" % (v, v))
                tester.set_write_selector(v)
                s = tester.get_write_selector()
                print("\tGet write selector: %s [%d]" % (s, s))
        elif action == "3":
            s = tester.get_number_format()
            print("\tGet number format: %s [%d]" % (s, s))
            for v in [NumberFormat.BYTE, NumberFormat.UINT16, NumberFormat.UINT32, NumberFormat.FLOAT]:
                print("\tSet number format to %s [%d]" % (v, v))
                tester.set_number_format(v)
                s = tester.get_number_format()
                print("\tGet number format: %s [%d]" % (s, s))
        elif action == "4":
            print("\tTurning leds off for 3 sec...")
            tester.control(ControlAction.LEDS_OFF)
        elif action == "5":
            print("\tTurning leds on (all red) for 3 sec...")
            tester.control(ControlAction.LEDS_ON_ALL_RED)
        elif action == "6":
            print("\tTurning leds on (all green) for 3 sec...")
            tester.control(ControlAction.LEDS_ON_ALL_GREEN)
        elif action == "7":
            print("\tTurning leds on (all yellow) for 3 sec...")
            tester.control(ControlAction.LEDS_ON_ALL_YELLOW)
        elif action == "8":
            print("\tTurning leds on (all default) for 3 sec...")
            tester.control(ControlAction.LEDS_ON_ALL_DEFAULT)
        elif action == "10":
            tester.set_number_format(NumberFormat.BYTE)
            tester.set_read_selector(ReadSelector.READ_MAX)
            v = tester.read_channel_value(NumberFormat.BYTE, 0)
            print("\tmax([%s]) is %d [0x%X]" % (NumberFormat.BYTE, v, v))
        elif action == "11":
            tester.set_number_format(NumberFormat.UINT16)
            tester.set_read_selector(ReadSelector.READ_MAX)
            v = tester.read_channel_value(NumberFormat.UINT16, 0)
            print("\tmax([%s]) is %d [0x%X]" % (NumberFormat.UINT16, v, v))
        elif action == "12":
            tester.set_number_format(NumberFormat.UINT32)
            tester.set_read_selector(ReadSelector.READ_MAX)
            v = tester.read_channel_value(NumberFormat.UINT32, 0)
            print("\tmax([%s]) is %d [0x%X]" % (NumberFormat.UINT32, v, v))
        elif action == "13":
            tester.set_number_format(NumberFormat.FLOAT)
            tester.set_read_selector(ReadSelector.READ_MAX)
            v = tester.read_channel_value(NumberFormat.FLOAT, 0)
            print("\tmax([%s]) is %e" % (NumberFormat.FLOAT, v))
        elif action == "14":
            tester.set_number_format(NumberFormat.FLOAT)
            tester.set_read_selector(ReadSelector.READ_MIN)
            v = tester.read_channel_value(NumberFormat.FLOAT, 0)
            print("\tmin([%s]) is %e" % (NumberFormat.FLOAT, v))
        elif action == "15":
            tester.set_number_format(NumberFormat.BYTE)
            tester.set_read_selector(ReadSelector.READ_CAFEBABE)
            v = tester.read_channel_value(NumberFormat.BYTE, 0)
            print("\tbyte(%s) is 0x%s [0x%x]" % (ReadSelector.READ_CAFEBABE, v, v))
        elif action == "16":
            tester.set_number_format(NumberFormat.UINT16)
            tester.set_read_selector(ReadSelector.READ_CAFEBABE)
            v = tester.read_channel_value(NumberFormat.UINT16, 0)
            print("\tuint16(%s) is 0x%s [0x%x]" % (ReadSelector.READ_CAFEBABE, v, v))
        elif action == "17":
            tester.set_number_format(NumberFormat.UINT32)
            tester.set_read_selector(ReadSelector.READ_CAFEBABE)
            v = tester.read_channel_value(NumberFormat.UINT32, 0)
            print("\tuint32([%s]) is 0x%s [0x%x]" % (ReadSelector.READ_CAFEBABE, v, v))
        elif action == "20":
            tester.set_number_format(NumberFormat.UINT16)
            tester.set_read_selector(ReadSelector.READ_POT)
            v = tester.read_channel_value(NumberFormat.UINT16, 0)
            print("\tPotentiometer[0] is %d" % v)
        elif action == "21":
            tester.set_number_format(NumberFormat.FLOAT)
            tester.set_read_selector(ReadSelector.READ_CV)
            v = tester.read_channel_value(NumberFormat.FLOAT, 0)
            print("\tCV uncalibrated[0] is %d" % v)
        elif action == "22":
            tester.set_number_format(NumberFormat.BYTE)
            tester.set_read_selector(ReadSelector.READ_SWITCH)
            v = tester.read_channel_value(NumberFormat.BYTE, 0)
            print("\tSwitch[0] is %d" % v)
        elif action == "30":
            print("\twriting %d 0x%x" % (0xCAFEBABE, 0xCAFEBABE))
            tester.set_number_format(NumberFormat.UINT32)
            tester.write_data(NumberFormat.UINT32, 0xCAFEBABE)

            tester.set_read_selector(ReadSelector.READ_WRITE_REGISTER)
            v = tester.read_channel_value(NumberFormat.UINT32, 0)
            print("\twrite_buffer now contains:  %d [0x%x] " % (v, v))
        elif action == "33":
            tester.set_number_format(NumberFormat.FLOAT)
            tester.set_read_selector(ReadSelector.READ_CALIBRATION_ADC_OFFSET)
            offset = tester.read_channel_value(NumberFormat.FLOAT, 0)
            tester.set_read_selector(ReadSelector.READ_CALIBRATION_ADC_SCALE)
            scale = tester.read_channel_value(NumberFormat.FLOAT, 0)
            print("\tADC[FM] calibration: %s, %s" % (offset, scale))
        elif action == "40":
            tester.set_list_selector(ListSelector.LIST_MANUFACTURER)
            n = tester.get_list_length()
            print("\tManufacturer list length %d" % n)
        elif action == "41":
            tester.set_list_selector(ListSelector.LIST_POT)
            n = tester.get_list_length()
            print("\tPot list length %d" % n)
        elif action == "42":
            values = tester.get_list(ListSelector.LIST_MANUFACTURER)
            print("\tManufacturer: %s" % values[0])
        elif action == "43":
            values = tester.get_list(ListSelector.LIST_POT)
            print("\tList of pots: %s" % values)
        elif action == "d":
            tester.print_calibration_data_json()
        print("")


def prompt_save_path_if_needed(args, prompt_txt):
    if args.path is None:
        path = input(prompt(prompt_txt))
        if path == "":
            return False
        elif path == "q":
            sys.exit(0)
        else:
            args.path = path

    p = pathlib.Path(args.path)
    if p.exists():
        print_attention("Are you sure to overwrite file at %s ?" % args.path)
        answer = input(prompt("Overwrite (y/n/q): ")).lower()
        if answer == "q":
            sys.exit(0)
        if answer != "y":
            return False
    return True


def save_mode(args, tester):
    global isSaveMode
    isSaveMode = True
    while not prompt_save_path_if_needed(args, "Save calibration data to path: "):
        continue

    calibration_data = tester.calibration_data_dump([NumberFormat.UINT32, NumberFormat.FLOAT])

    connected_module = Module()
    connected_module.fetch_info(tester)

    creation_date = datetime.datetime.utcnow().isoformat() + 'Z'
    script_name = sys.argv[0]

    d = {
        'creation_date': creation_date,
        'producer': script_name,
        'calibration_data': calibration_data,
        'module_info': connected_module.info_dict()
    }

    json_data = json.dumps(d, indent=4)
    p = pathlib.Path(args.path)
    p.write_text(json_data)
    print("Successfully saved JSON file %s" % args.path)


def load_mode(args, tester):
    global is_open_mode
    is_open_mode = True
    while True:
        if args.path is None:
            path = input(prompt("Load calibration data from path: "))
            if path == "":
                continue
            elif path == "q":
                sys.exit(0)
            else:
                args.path = path
        else:
            print("Loading JSON file %s" % args.path)

        p = pathlib.Path(args.path)
        if p.exists():
            with p.open(encoding="UTF-8") as source:
                d = json.load(source)

                if "calibration_data" in d:
                    d = d["calibration_data"]

                json_data = json.dumps(d, indent=4)
                print("%s\n" % json_data)

                print_attention("Are you sure to load the above calibration settings into RAM?")
                answer = input(prompt("Update RAM (y/n): ")).lower()
                if answer != "y":
                    args.path = None
                    continue

                tester.restore_calibration_data(d)

                print("Successfully updated MCU RAM to:")
                tester.print_calibration_data_json()

                print_attention("Are you sure to flash the above RAM content?")
                answer = input(prompt("Flash (y/n): ")).lower()
                if answer != "y":
                    args.path = None
                    continue

                tester.control(ControlAction.FLASH_SETTINGS)

            sys.exit(0)
        else:
            print_error("File does not exist at path %s" % args.path)


def write_dac_lut_mode(args, tester):
    global isSaveMode
    isSaveMode = True

    meter = prepare_voltmeter(args)

    while not prompt_save_path_if_needed(args, "Save DAC lookup table CSV: "):
        continue

    build_dac_lut(tester, meter, args)

    print("Successfully saved CSV file %s" % args.path)


# -----------------------------------------------------------------------------
# MARK: -
def main():
    args = parse_args()
    if args is None:
        sys.exit(1)
    # print(args)

    uart = serial.Serial(args.port, args.baudRate, timeout=10)
    print("Using UART %s at %d baud\n" % (uart.name, uart.baudrate))

    module = Tides()
    tester = Tester(module, uart, args.verbose)

    if args.mode == "info":
        info_mode(tester)
    elif args.mode == "inputs":
        inputs_mode(args, tester)
    elif args.mode == "leds":
        led_mode(args, tester)
    elif args.mode == "dac":
        dac_mode(args, tester)
    elif args.mode == "adc":
        adc_mode(args, tester)
    elif args.mode == "dev":
        dev_mode(args, tester)
    elif args.mode == "save":
        save_mode(args, tester)
    elif args.mode == "load":
        load_mode(args, tester)
    elif args.mode == "lut-dac":
        write_dac_lut_mode(args, tester)
    uart.close()


main()
