#! /usr/bin/env python3 -u

# -----------------------------------------------------------------------------
# Copyright 2021 Martin Jan Köhler
# See http://creativecommons.org/licenses/MIT/ for more information.
# -----------------------------------------------------------------------------

import argparse
import os
import re
import readline
import serial
import statistics
import sys
import time
import pyvisa

def encode(str):
	return str.encode("unicode_escape").decode("ascii")

def parse_args():
    parser = argparse.ArgumentParser(description="Voltmeter Console")
    parser.add_argument('--device', dest='port', action='store', type=str, default="ASRL4::INSTR", required=False, help='VISA port (default ASRL4::INSTR)')
    parser.add_argument('--in-termination', dest='read_termination', action='store', type=str, default='\r', required=False, help='SCPI RX command termination (default %s)' % encode('\r'))
    parser.add_argument('--out-termination', dest='write_termination', action='store', type=str, default='\r\n', required=False, help='SCPI TX command termination (default %s)' % encode('\r\n'))
    parser.add_argument('--baudrate', dest='baudRate', type=int, default=9600, help='UART baud rate')
    parser.add_argument('--debug', dest='debug', action='store_true', default=False, help='Enable Debug logging')

    args = parser.parse_args()
    return args

class Keithley2015Multimeter:
	def __init__(self, instrument):
		self.inst = instrument
	
	def close(self):
		self.inst.close()

	def command(self, str):
		result = self.inst.write(str)
		return result
	
	def query(self, str):
		result = self.inst.query(str).rstrip()
		return result
	
	def voltage(self, num_readings):
		voltages = []
		v = self.query("func 'volt:dc';:read?")
		for i in range(0, num_readings):
			if i >= 1:
				time.sleep(0.1)
			v = self.query("func 'volt:dc';:read?")
			try:
				v = float(v)
				voltages.append(v)
				
				avg = round(statistics.mean(voltages), 5)
				median = round(statistics.median(voltages), 5)
				# print("avg: %s, median: %s" % (avg, median))
			except ValueError as e:
				print("ERROR: %s" % e)
				continue
			except pyvisa.VisaIOError as e:
				print("ERROR: %s" % e)
				continue
		missing_items = num_readings - len(voltages)
		if missing_items > 0:
			voltages += self.voltage(missing_items)
		return voltages
	
	def reset(self):
		# Reset the SENSe1 subsystem settings, along with the trigger model,
		# each READ? will cause one trigger
		self.command("*RST")
		self.command("*CLS")
		got_timeout = False
		while not got_timeout:
			try:
				self.inst.read()
			except pyvisa.VisaIOError as e:
				got_timeout = True
	
	def identification(self):
		return self.query("*IDN?")
	
	def systemVersion(self):
		return self.query(":SYSTem:VERSion?")

def main():
	readline.parse_and_bind('tab: complete')

	args = parse_args()

	if args.debug:
		pyvisa.log_to_screen()
	
	resourceManager = pyvisa.ResourceManager()
	instrument = resourceManager.open_resource(
		args.port,
		read_termination=args.read_termination,
		write_termination=args.write_termination,
		baud_rate=args.baudRate
	)
	instrument.clear()
	
	meter = Keithley2015Multimeter(instrument)
	meter.reset()
	
	print("Detected device id: %s" % meter.identification())
	print("---------------------------------------------------------------------")
	time.sleep(1)

	voltages = meter.voltage(10)

	meter.close()

if __name__ == "__main__":
	main()
